import time
import asyncio
import logging
import datetime as dt
from contextlib import AsyncExitStack
from types import TracebackType
from typing import ClassVar, Self
from collections import Counter
from concurrent.futures import ProcessPoolExecutor

from uptime.frigate_config import FrigateConfig
from uptime.notify import NtfyNotifier
from uptime.ping import video_ping

logger = logging.getLogger(__name__)


class CameraMonitor:
    _FRIGATE_CONFIG_REFRESH_INTERVAL: ClassVar[dt.timedelta] = dt.timedelta(minutes=60)

    def __init__(self, ping_interval: dt.timedelta = dt.timedelta(seconds=60), consecutive_down_threshold: int = 3):
        self._ping_interval = ping_interval
        self._consecutive_down_threshold = consecutive_down_threshold
        self._camera_downtime = Counter()
        self._frigate_config = FrigateConfig()
        self._ntfy_notifier = NtfyNotifier()
        self._process_pool = ProcessPoolExecutor()

    async def _on_camera_up(self, camera: str) -> None:
        logger.info(f"Camera {camera} is back online")
        await self._ntfy_notifier.send_notification(camera, True)

    async def _on_camera_down(self, camera: str) -> None:
        await self._frigate_config.refresh()
        if camera not in self._frigate_config.active_cameras:
            return

        logger.info(f"Camera {camera} is down")
        await self._ntfy_notifier.send_notification(camera, False)

    async def _on_camera_ping(self, camera: str, success: bool) -> None:
        if success:
            if self._camera_downtime[camera] >= self._consecutive_down_threshold:
                await self._on_camera_up(camera)
            self._camera_downtime[camera] = 0
        else:
            if self._camera_downtime[camera] == self._consecutive_down_threshold:
                await self._on_camera_down(camera)
            self._camera_downtime[camera] += 1

    async def _ping_all_cameras(self) -> None:
        async def video_ping_async(url: str) -> bool:
            event_loop = asyncio.get_running_loop()
            return await event_loop.run_in_executor(
                self._process_pool,
                video_ping,
                url,
            )

        ping_tasks = {}
        async with asyncio.TaskGroup() as tg:
            event_loop = asyncio.get_running_loop()
            for camera, url in self._frigate_config.active_camera_urls.items():
                ping_tasks[camera] = tg.create_task(video_ping_async(url))

        for camera, task in ping_tasks.items():
            await self._on_camera_ping(camera, task.result())

    async def run(self) -> None:
        await self._frigate_config.refresh()
        last_refresh = time.time()
        while True:
            start_time = time.time()
            if time.time() - last_refresh >= self._FRIGATE_CONFIG_REFRESH_INTERVAL.total_seconds():
                logger.info("Refreshing Frigate config...")
                await self._frigate_config.refresh()
                last_refresh = time.time()
            await self._ping_all_cameras()

            sleep_duration = self._ping_interval.total_seconds() - (time.time() - start_time)
            logger.debug(f"Sleeping for {sleep_duration} seconds...")
            await asyncio.sleep(sleep_duration)
